import homekeeper.common
import logging
import os
import shutil

common = homekeeper.common


def symlink(config, source, target):
    """Removes a target symlink/file/directory before replacing it with
    symlink. Also creates the parent directory if it does not exist.

    Args:
        config: Homekeeper configuration.
        source: Original source of the symlink.
        target: Path of symlink target, can be file or directory.
    """
    dirname = os.path.dirname(target)
    if not os.path.exists(dirname):
        common.makedirs(dirname)
    if source == target:
        logging.info('skipping %s; source and target are the same', source)
        return
    if os.path.exists(target) and not config.overwrite:
        logging.info('skipping %s; will not overwrite', target)
        return
    remove(target)
    os.symlink(source, target)
    logging.info('symlinked %s -> %s', source, target)


def restore(unused_config, source, target): # pylint: disable=unused-argument
    """Restores a symlink to its target. Afterwards, the target will no
    longer be a symlink.

    Args:
        config: Homekeeper configuration.
        source: Original source of the symlink.
        target: Path of symlink target, can be file or directory.
    """
    if source == target:
        logging.info('skipping %s; source and target are the same', source)
        return
    if not os.path.islink(target):
        logging.info('skipping %s; resource is not a link', target)
        return
    if os.readlink(target) != source:
        logging.info('skipping %s; symlink target is wrong', target)
        return
    remove(target)
    if os.path.isfile(source):
        shutil.copy(source, target)
        logging.info('restored file %s -> %s', source, target)
    elif os.path.isdir(source):
        shutil.copytree(source, target, symlinks=True)
        logging.info('restored directory %s -> %s', source, target)
    else:
        logging.info('skipping %s; not a file or directory', target)


def create_symlinks(config, source_directory, target_directory):
    """Symlinks files from the source directory to the target directory.

    For example, suppose that your `source_directory` is your dotfiles
    directory and contains a file named '.vimrc'.  If the `target_directory`
    is your home directory, then this is the result:

        $HOME/.vimrc -> $HOME/dotfiles/.vimrc

    The existing $HOME/.vimrc will be removed.

    Args:
        config: Homekeeper configuration.
        source_directory: The source directory where your dotfiles are.
        target_directory: The target directory for symlinking.
    """
    process_directories(config, source_directory, target_directory, symlink)


def create_symlinks_from_base(config):
    if not config.override:
        return
    create_symlinks(config, config.base_directory, config.home)


def create_symlinks_from_dotfiles(config):
    create_symlinks(config, config.dotfiles_directory, config.home)


def restore_symlinks(config, source_directory, target_directory):
    """Realizes the symlinks files in the source directory that have been
    symlinked to the target directory.

    For example, suppose your home directory contains:

        $HOME/.vimrc -> $HOME/dotfiles/.vimrc

    Then the result will be:

        $HOME/.vimrc

    The $HOME/.vimrc file will contain to contains of $HOME/dotfiles/.vimrc.
    Any symlinks in the source directory that do not point to the target
    directory will not be restored.

    Args:
        config: Homekeeper configuration.
        source_directory: The source directory where your dotfiles are.
        target_directory: The target directory for symlinking.
    """
    process_directories(config, source_directory, target_directory, restore)


def restore_symlinks_from_base(config):
    if not config.override:
        return
    restore_symlinks(config, config.base_directory, config.home)


def restore_symlinks_from_dotfiles(config):
    restore_symlinks(config, config.dotfiles_directory, config.home)


def cleanup_symlinks(directory):
    """Removes broken symlinks from a directory.

    Args:
        directory: The directory to look for broken symlinks.
    """
    logging.info('removing broken symlinks in %s', directory)
    for item in os.listdir(directory):
        pathname = os.path.join(directory, item)
        if not os.path.islink(pathname):
            logging.debug('skipping %s; not a symlink', pathname)
            continue
        if os.path.exists(os.readlink(pathname)):
            logging.debug('skipping %s; not a broken symlink', pathname)
            continue
        logging.info('removing broken link: %s', pathname)
        os.unlink(pathname)
    logging.info('finished removing symlinks in %s', directory)


def remove(target):
    if os.path.islink(target):
        os.unlink(target)
        logging.debug('removed symlink %s', target)
    if os.path.isfile(target):
        os.remove(target)
        logging.debug('removed file %s', target)
    if os.path.isdir(target):
        shutil.rmtree(target)
        logging.debug('removed directory %s', target)


def process_directories(config, source_directory, target_directory, process):
    if not os.path.isdir(source_directory):
        logging.info('dotfiles directory not found: %s', source_directory)
        return
    if source_directory == target_directory:
        logging.error('source and target directory are the same')
        return
    logging.info('processing files in %s', source_directory)
    with common.cd(source_directory):
        # Excludes must include included files; or else they will be included
        # twice in the second for loop.
        includes = [p for p in config.includes
                    if is_valid_include(p, source_directory)]
        excludes = [firstpart(relativize(p, source_directory))
                    for p in includes] + config.excludes
        logging.debug('excluding items: %s', excludes)
        for pathname in os.listdir('.'):
            basename = os.path.basename(pathname)
            source = os.path.join(source_directory, basename)
            target = os.path.join(target_directory, basename)
            if basename in excludes:
                logging.debug('skipping %s; resource is excluded', basename)
                continue
            logging.debug('processing %s -> %s', source, target)
            process(config, source, target)
        for pathname in includes:
            relpath = relativize(pathname, source_directory)
            source = os.path.join(source_directory, relpath)
            target = os.path.join(target_directory, relpath)
            logging.debug('processing %s -> %s', source, target)
            process(config, source, target)


def relativize(pathname, commonprefix):
    abspath = os.path.abspath(pathname)
    return os.path.relpath(abspath, commonprefix)


def firstpart(pathname):
    head, tail = os.path.split(pathname)
    if head is os.sep:
        return head
    elif not head:
        return tail
    else:
        return firstpart(head)


def is_valid_include(pathname, commonprefix):
    logging.debug('checking validity of include: %s', pathname)
    relpath = relativize(pathname, commonprefix)
    abspath = os.path.join(commonprefix, relpath)
    basename = os.path.basename(abspath)
    if '.' == basename:
        logging.debug('invalid because basename was .: %s', abspath)
        return False
    if '..' in abspath:
        logging.debug('invalid due to ./.. in path: %s', abspath)
        return False
    if not os.path.exists(abspath):
        logging.debug('invalid due to non-existing path: %s', abspath)
        return False
    logging.debug('include was valid: %s', abspath)
    return True



import click
import homekeeper
import logging


@click.command(short_help='Removes broken symlinks only.')
@click.pass_context
def cleanup(ctx):
    h = create_homekeeper(ctx.obj)
    h.cleanup_symlinks = True
    h.overwrite = False
    h.cleanup()


@click.command(short_help='Set dotfiles directory to current directory.')
@click.pass_context
def init(ctx):
    h = create_homekeeper(ctx.obj)
    h.cleanup_symlinks = False
    h.overwrite = False
    h.init()


@click.command(short_help='Symlinks dotfiles to your home directory.')
@click.pass_context
def keep(ctx):
    h = create_homekeeper(ctx.obj)
    h.keep()


@click.command(short_help='Restores dotfiles and replacing symlinks.')
@click.pass_context
def unkeep(ctx):
    h = create_homekeeper(ctx.obj)
    h.unkeep()


@click.command(short_help='Displays the version and then exits.')
@click.pass_context
def version(ctx): # pylint: disable=unused-argument
    click.echo('Homekeeper version %s' % homekeeper.__version__)


@click.group()
@click.option('--cleanup-symlinks/--no-cleanup-symlinks', default=True,
              is_flag=True, help='Removes broken symlinks (default true).')
@click.option('--overwrite/--no-overwrite', default=True, is_flag=True,
              help='Overwrite existing files or symlinks (default true).')
@click.option('--config-path', metavar='FILE', default=None,
              help='Path to .homekeeper.json config file.')
@click.option('--debug/--no-debug', default=False, is_flag=True,
              help='Enables debug output (default false).')
@click.pass_context
def main(ctx, cleanup_symlinks, overwrite, config_path, debug):
    if debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(format='%(message)s', level=logging.INFO)
    ctx.obj = {
        'cleanup_symlinks': cleanup_symlinks,
        'config_path': config_path,
        'overwrite': overwrite,
    }


def create_homekeeper(args):
    h = homekeeper.Homekeeper(config_path=args['config_path'])
    h.cleanup_symlinks = args['cleanup_symlinks']
    h.overwrite = args['overwrite']
    return h


main.add_command(cleanup)
main.add_command(init)
main.add_command(keep)
main.add_command(unkeep)
main.add_command(version)



import contextlib
import pyfakefs.fake_filesystem
import pyfakefs.fake_filesystem_shutil
import homekeeper.common
import json
import logging
import mock
import pytest

fake_filesystem = pyfakefs.fake_filesystem
fake_filesystem_shutil = pyfakefs.fake_filesystem_shutil
logging.basicConfig(format='%(message)s', level=logging.DEBUG)


# pylint: disable=attribute-defined-outside-init
# pylint: disable=no-self-use
# pylint: disable=too-many-instance-attributes
class TestCase(object):
    def setup_method(self):
        self.fake_fs = fake_filesystem.FakeFilesystem()
        self.fake_fopen = fake_filesystem.FakeFileOpen(self.fake_fs)
        self.fake_os = fake_filesystem.FakeOsModule(self.fake_fs)
        self.fake_shutil = fake_filesystem_shutil.FakeShutilModule(self.fake_fs)
        self.patchers = []
        self.setup_os(self.fake_os)
        self.setup_configuration(self.fake_os)

    def setup_os(self, os):
        os.environ['HOME'] = os.path.join(os.sep, 'home', 'johndoe')
        self.home = os.getenv('HOME')
        self.setup_directory(self.home)

    def setup_configuration(self, os):
        with contextlib.nested(mock.patch('homekeeper.common.os', self.fake_os),
                               homekeeper.common.cd(self.home)):
            self.base_directory = os.path.abspath('base')
            self.dotfiles_directory = os.path.abspath('dotfiles')
            self.custom_directory = os.path.abspath('custom')
            self.setup_directory(self.base_directory)
            self.setup_directory(self.dotfiles_directory)
            self.setup_directory(self.custom_directory)
            self.setup_directory(self.custom_directory, 'base')
            self.setup_directory(self.custom_directory, 'dotfiles')
        data = json.dumps({
            'base_directory': self.base_directory,
            'dotfiles_directory': self.dotfiles_directory,
            'excludes': ['.git', '.gitignore'],
            'includes': [os.path.join(self.home, '.foo', 'foorc')],
        })
        self.setup_file(self.home, '.homekeeper.json', data=data)
        data = json.dumps({
            'base': os.path.join(self.custom_directory, 'base'),
            'directory': os.path.join(self.custom_directory, 'dotfiles'),
            'excludes': ['.git'],
            'cherrypicks': [os.path.join(self.home, '.bar', 'barrc')],
        })
        self.setup_file(self.home, 'custom', '.homekeeper.json', data=data)

    def teardown_method(self):
        for patcher in self.patchers:
            patcher.stop()
        del self.fake_fs

    @pytest.fixture
    def os(self):
        return self.fake_os

    def patch(self, module):
        self._patch(module, 'fopen', self.fake_fopen)
        self._patch(module, 'os', self.fake_os)
        self._patch(module, 'shutil', self.fake_shutil)

    def read_file(self, *args):
        return self._read_file(self.fake_os, self.fake_fopen, *args)

    def setup_file(self, *args, **kwargs):
        return self._setup_file(self.fake_os, args, kwargs)

    def setup_directory(self, *args):
        return self._setup_directory(self.fake_os, args)

    def _read_file(self, os, fopen, *args):
        filename = os.path.join(*args)
        with fopen(filename, 'r') as f:
            return f.read()

    def _setup_file(self, os, args, kwargs):
        filename = os.path.join(*args) # pylint: disable=star-args
        dirname = os.path.dirname(filename)
        self.setup_directory(dirname)
        contents = '' if ('data' not in kwargs) else kwargs['data']
        if os.path.exists(filename):
            os.unlink(filename)
        self.fake_fs.CreateFile(filename, contents=contents)
        return filename

    def _setup_directory(self, os, args):
        dirname = os.path.join(*args) # pylint: disable=star-args
        if dirname == '' or dirname == os.sep:
            return
        with mock.patch('homekeeper.common.os', os):
            homekeeper.common.makedirs(dirname)
        return dirname

    def _patch(self, module, name, fake):
        try:
            patcher = mock.patch(module + '.' + name, fake)
            patcher.start()
            self.patchers.append(patcher)
        except AttributeError:
            logging.debug('can not patch: %s.%s', module, name)



import homekeeper.config
import homekeeper.core
import logging
import os
import shutil

__version__ = '4.0.2'
core = homekeeper.core


class Homekeeper(object):
    """Organizes and versions your dot files."""

    def __init__(self, config_path=None):
        self.config_path = config_path
        self.config = homekeeper.config.Config()
        self.config.load(self.config_path)

    def init(self):
        """Writes a configuration file with cwd as the dotfiles directory.
        Configuration file is written as JSON, and will be removed if it exists
        already.  If configuration already exists, the new dotfiles directory
        path will be merged into existing configuration.
        """
        dotfiles_directory = os.path.realpath(os.getcwd())
        logging.info('setting dotfiles directory to %s', dotfiles_directory)
        self.config.dotfiles_directory = dotfiles_directory
        self.config.save(self.config_path)

    def keep(self):
        """Symlinks all files and directories from your dotfiles directory into
        your home directory.
        """
        core.create_symlinks_from_base(self.config)
        core.create_symlinks_from_dotfiles(self.config)
        self.cleanup()

    def link(self):
        self.keep()

    def unkeep(self):
        """Restores all symlinks (inverse of link)."""
        core.restore_symlinks_from_base(self.config)
        core.restore_symlinks_from_dotfiles(self.config)
        self.cleanup()

    def restore(self):
        self.unkeep()

    def cleanup(self):
        """Cleans up symlinks in the home directory."""
        if not self.config.cleanup_symlinks:
            logging.info('skipping cleanup of broken symlinks')
            return
        core.cleanup_symlinks(self.config.home)

    @property
    def cleanup_symlinks(self):
        return self.config.cleanup_symlinks

    @property
    def overwrite(self):
        return self.config.overwrite

    @cleanup_symlinks.setter
    def cleanup_symlinks(self, value):
        self.config.cleanup_symlinks = value

    @overwrite.setter
    def overwrite(self, value):
        self.config.overwrite = value



import homekeeper
import homekeeper.common
import homekeeper.config
import homekeeper.test_case
import json

cd = homekeeper.common.cd

# pylint: disable=attribute-defined-outside-init
# pylint: disable=no-self-use
class TestHomekeeper(homekeeper.test_case.TestCase):
    def setup_method(self):
        super(TestHomekeeper, self).setup_method()
        self.patch('homekeeper')
        self.patch('homekeeper.common')
        self.patch('homekeeper.config')
        self.patch('homekeeper.core')
        self.setup_files()

    def setup_files(self):
        with cd(self.base_directory):
            self.setup_file('.bash_aliases', data='base')
            self.setup_file('.bash_local', data='base')
            self.setup_file('.bash_profile')
            self.setup_file('.git')
            self.setup_file('.gitconfig')
            self.setup_file('.gitignore')
            self.setup_file('.vimrc')
            self.setup_directory('.tmux')
            self.setup_directory('.tmux', 'base')
            self.setup_directory('.tmuxp')
            self.setup_directory('.tmuxp', 'base')
            self.setup_directory('.vim')
        with cd(self.dotfiles_directory):
            self.setup_file('.bash_aliases', data='dotfiles')
            self.setup_file('.bash_local', data='dotfiles')
            self.setup_file('.xbindkeysrc', data='dotfiles')
            self.setup_directory('.tmux')
            self.setup_directory('.tmuxp')
            self.setup_directory('bin')
        with cd(self.custom_directory):
            self.setup_directory('base')
            self.setup_directory('dotfiles')

    def test_setters_override_config(self):
        h = homekeeper.Homekeeper()
        assert h.config.overwrite
        assert h.config.cleanup_symlinks
        h.overwrite = False
        h.cleanup_symlinks = False
        assert not h.config.overwrite
        assert not h.config.cleanup_symlinks

    def test_init_saves_config(self, os):
        with cd(self.custom_directory):
            config = os.path.join(self.custom_directory, '.homekeeper.json')
            homekeeper.Homekeeper(config_path=config).init()
            data = json.loads(self.read_file(config))
            assert self.custom_directory == data['dotfiles_directory']

    def test_init_with_default_config_path(self):
        h = homekeeper.Homekeeper()
        assert h.config.base_directory == self.base_directory
        assert h.config.dotfiles_directory == self.dotfiles_directory
        assert '.git' in h.config.excludes

    def test_keep_overrides_base_files(self, os):
        h = homekeeper.Homekeeper()
        h.keep()
        base_items = set(os.listdir(self.base_directory))
        dotfiles_items = set(os.listdir(self.dotfiles_directory))
        for item in os.listdir(self.home):
            assert item not in h.config.excludes
            if item not in base_items and item not in dotfiles_items:
                continue
            link = os.path.join(self.home, item)
            target_directory = (self.dotfiles_directory if
                                item in dotfiles_items else self.base_directory)
            assert os.path.islink(link)
            assert os.path.join(target_directory, item) == os.readlink(link)

    def test_unkeep_restores_files(self, os):
        self.setup_file(self.home, '.unrelatedrc')
        os.symlink(os.path.join(self.home, '.unrelatedrc'),
                   os.path.join(self.home, '.link-to-unrelatedrc'))
        os.symlink(os.path.join(self.base_directory, '.vimrc'),
                   os.path.join(self.home, '.vimrc'))
        os.symlink(os.path.join(self.base_directory, '.vim'),
                   os.path.join(self.home, '.vim'))
        os.symlink(os.path.join(self.dotfiles_directory, '.bash_local'),
                   os.path.join(self.home, '.bash_local'))
        os.symlink(os.path.join(self.dotfiles_directory, '.tmux'),
                   os.path.join(self.home, '.tmux'))
        h = homekeeper.Homekeeper()
        h.unkeep()
        assert os.path.exists(os.path.join(self.home, '.unrelatedrc'))
        assert os.path.islink(os.path.join(self.home, '.link-to-unrelatedrc'))
        assert not os.path.islink(os.path.join(self.home, '.unrelatedrc'))
        assert not os.path.islink(os.path.join(self.home, '.tmux'))
        assert not os.path.islink(os.path.join(self.home, '.vim'))
        assert not os.path.islink(os.path.join(self.home, '.bash_local'))
        assert not os.path.islink(os.path.join(self.home, '.vimrc'))
        assert 'dotfiles' == self.read_file(self.home, '.bash_local')

    def test_cleanup_symlinks(self, os):
        self.setup_file(self.home, '.missingrc')
        os.symlink(os.path.join(self.home, '.missingrc'),
                   os.path.join(self.home, '.brokenrc'))
        os.unlink(os.path.join(self.home, '.missingrc'))
        os.symlink(os.path.join(self.base_directory, '.vimrc'),
                   os.path.join(self.home, '.vimrc'))
        assert os.path.islink(os.path.join(self.home, '.brokenrc'))
        assert os.path.islink(os.path.join(self.home, '.vimrc'))
        h = homekeeper.Homekeeper()
        h.cleanup_symlinks = True
        h.cleanup()
        assert not os.path.exists(os.path.join(self.home, '.brokenrc'))
        assert os.path.islink(os.path.join(self.home, '.vimrc'))



import click.testing
import homekeeper
import homekeeper.cli
import homekeeper.test_case
import mock


# pylint: disable=attribute-defined-outside-init
# pylint: disable=maybe-no-member
class TestCli(homekeeper.test_case.TestCase):
    def setup_method(self):
        super(TestCli, self).setup_method()
        self.runner = click.testing.CliRunner()
        self.mock_instance = mock.MagicMock()
        self.patcher = mock.patch('homekeeper.Homekeeper')
        self.mock_class = self.patcher.start()
        self.mock_class.return_value = self.mock_instance

    def teardown_method(self):
        self.patcher.stop()

    def run(self, *args, **kwargs):
        result = self.runner.invoke(homekeeper.cli.main, args)
        assert result.exit_code == 0
        if 'output' in kwargs:
            assert result.output == kwargs['output']
        else:
            assert result.output == ''

    def test_cleanup(self):
        self.run('cleanup')
        self.mock_class.cleanup.assert_called_once()
        assert self.mock_instance.cleanup_symlinks == True
        assert self.mock_instance.overwrite == False
        self.mock_instance.cleanup.assert_called_with()

    def test_cleanup_ignores_no_cleanup_and_overwrite(self):
        self.run('--no-cleanup-symlinks', '--overwrite', 'cleanup')
        self.mock_class.cleanup.assert_called_once()
        assert self.mock_instance.cleanup_symlinks == True
        assert self.mock_instance.overwrite == False
        self.mock_instance.cleanup.assert_called_with()

    def test_init_with_custom_config(self, os):
        dotfiles_directory = os.getcwd()
        config_path = os.path.join(dotfiles_directory, '.config.json')
        self.run('--config-path', config_path, 'init')
        self.mock_class.assert_called_once_with(config_path=config_path)
        self.mock_class.init.assert_called_once()
        assert self.mock_instance.cleanup_symlinks == False
        assert self.mock_instance.overwrite == False
        self.mock_instance.init.assert_called_with()

    def test_init_cannot_cleanup_or_overwrite(self):
        self.run('--cleanup-symlinks', '--overwrite', 'init')
        self.mock_class.assert_called_once_with(config_path=None)
        self.mock_class.init.assert_called_once()
        assert self.mock_instance.cleanup_symlinks == False
        assert self.mock_instance.overwrite == False
        self.mock_instance.init.assert_called_with()

    def test_keep_with_default_options(self):
        self.run('keep')
        self.mock_class.assert_called_once_with(config_path=None)
        self.mock_class.keep.assert_called_once()
        assert self.mock_instance.cleanup_symlinks == True
        assert self.mock_instance.overwrite == True
        self.mock_instance.keep.assert_called_with()

    def test_keep_with_custom_options(self):
        self.run('--no-cleanup-symlinks', '--no-overwrite', 'keep')
        self.mock_class.assert_called_once_with(config_path=None)
        self.mock_class.keep.assert_called_once()
        assert self.mock_instance.cleanup_symlinks == False
        assert self.mock_instance.overwrite == False
        self.mock_instance.keep.assert_called_with()

    def test_unkeep_with_default_options(self):
        self.run('unkeep')
        self.mock_class.assert_called_once_with(config_path=None)
        self.mock_class.keep.assert_called_once()
        assert self.mock_instance.cleanup_symlinks == True
        assert self.mock_instance.overwrite == True
        self.mock_instance.unkeep.assert_called_with()

    def test_unkeep_with_custom_options(self):
        self.run('--no-cleanup-symlinks', '--no-overwrite', 'unkeep')
        self.mock_class.assert_called_once_with(config_path=None)
        self.mock_class.keep.assert_called_once()
        assert self.mock_instance.cleanup_symlinks == False
        assert self.mock_instance.overwrite == False
        self.mock_instance.unkeep.assert_called_with()

    def test_version(self):
        output = 'Homekeeper version %s\n' % homekeeper.__version__
        self.run('version', output=output)
        self.mock_class.assert_not_called()



import homekeeper.config
import homekeeper.test_case
import json


# pylint: disable=attribute-defined-outside-init
class TestConfig(homekeeper.test_case.TestCase):
    def setup_method(self):
        super(TestConfig, self).setup_method()
        self.patch('homekeeper.common')
        self.patch('homekeeper.config')
        self.config = homekeeper.config.Config()
        self.config_path = self.fake_os.path.join(self.home, '.homekeeper.json')

    def test_load(self, os):
        self.config.load(self.config_path)
        excludes = ['.git', '.gitignore']
        includes = [os.path.join(self.home, '.foo', 'foorc')]
        assert self.base_directory == self.config.base_directory
        assert self.dotfiles_directory == self.config.dotfiles_directory
        assert excludes == self.config.excludes
        assert includes == self.config.includes
        assert self.config.override

    def test_load_with_defaults(self):
        self.setup_file(self.config_path, data=json.dumps({}))
        self.config.load(self.config_path)
        assert not self.config.base_directory
        assert self.dotfiles_directory == self.config.dotfiles_directory
        assert [] == self.config.excludes
        assert [] == self.config.includes
        assert not self.config.override

    def test_load_and_save_old_config(self, os):
        config_path = os.path.join(self.home, 'custom', '.homekeeper.json')
        base_directory = os.path.join(self.home, 'custom', 'base')
        dotfiles_directory = os.path.join(self.home, 'custom', 'dotfiles')
        excludes = ['.git']
        includes = [os.path.join(self.home, '.bar', 'barrc')]
        self.config.load(config_path)
        assert base_directory == self.config.base_directory
        assert dotfiles_directory == self.config.dotfiles_directory
        assert excludes == self.config.excludes
        assert includes == self.config.includes
        assert self.config.override
        self.config.save(config_path)
        data = json.loads(self.read_file(config_path))
        assert data['base_directory'] == base_directory
        assert data['dotfiles_directory'] == dotfiles_directory
        assert data['excludes'] == excludes
        assert data['includes'] == includes
        assert 'override' not in data
        assert 'cherrypicks' not in data
        assert 'base' not in data
        assert 'directory' not in data

    def test_save(self, os):
        self.setup_directory(os.path.dirname(self.config_path))
        self.config.base_directory = None
        self.config.dotfiles_directory = os.path.join(self.home, 'custom')
        self.config.excludes = ['.idea']
        self.config.override = False
        self.config.save(self.config_path)
        data = json.loads(self.read_file(self.config_path))
        assert data['base_directory'] == self.config.base_directory
        assert data['dotfiles_directory'] == self.config.dotfiles_directory
        assert data['excludes'] == self.config.excludes
        assert 'override' not in data
        assert 'cherrypicks' not in data
        assert 'base' not in data
        assert 'directory' not in data



import homekeeper.common
import logging
import json
import os

common = homekeeper.common


class Config(object):
    """Representation of the homekeeper configuration file."""

    # pylint: disable=too-many-instance-attributes
    def __init__(self):
        self.home = os.getenv('HOME')
        self.base_directory = None
        self.dotfiles_directory = os.path.join(self.home, 'dotfiles')
        self.includes = []
        self.excludes = ['.git', '.gitignore', 'LICENSE', 'README.md']
        self.cleanup_symlinks = True
        self.overwrite = True
        self.override = False
        self.default_path = os.path.join(self.home, '.homekeeper.json')

    def load(self, pathname=None):
        config_path = pathname or self.default_path
        with common.fopen(config_path, 'r') as f:
            data = json.loads(f.read())
            if 'base' in data:
                self.base_directory = data['base']
            if 'base_directory' in data:
                self.base_directory = data['base_directory']
            if 'directory' in data:
                self.dotfiles_directory = data['directory']
            if 'dotfiles_directory' in data:
                self.dotfiles_directory = data['dotfiles_directory']
            if 'cherrypicks' in data:
                self.includes = data['cherrypicks']
            if 'includes' in data:
                self.includes = data['includes']
            if 'excludes' in data:
                self.excludes = data['excludes']
            else:
                self.excludes = []
            self.override = self.base_directory is not None
            logging.info('loaded configuration from %s', config_path)

    def save(self, pathname):
        with common.fopen(pathname, 'w') as f:
            data = {
                'base_directory': self.base_directory,
                'dotfiles_directory': self.dotfiles_directory,
                'includes': self.includes,
                'excludes': self.excludes,
            }
            f.write(json.dumps(data, sort_keys=True, indent=4))
            logging.info('saved configuration to %s', pathname)



import errno
import os


class cd(object): # pylint: disable=too-few-public-methods
    "Use with the `with` keyword to change directory."""
    def __init__(self, pathname):
        self.pathname = pathname
        self.saved_pathname = None

    def __enter__(self):
        self.saved_pathname = os.getcwd()
        os.chdir(self.pathname)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.saved_pathname)


def fopen(*args, **kwargs):
    """Alias for __builtin__.open. This exists so tests may mock this function
    without overriding open for all modules.
    """
    return open(*args, **kwargs)


def makedirs(dirname):
    """Creates a directory safely. Does not raise an error if the directory
    exists.
    """
    try:
        os.makedirs(dirname)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(dirname):
            pass
        else:
            raise



import homekeeper.common
import homekeeper.config
import homekeeper.core
import homekeeper.test_case

cd = homekeeper.common.cd
core = homekeeper.core


# pylint: disable=attribute-defined-outside-init
# pylint: disable=no-self-use
# pylint: disable=too-many-public-methods
class TestCore(homekeeper.test_case.TestCase):
    def setup_method(self):
        super(TestCore, self).setup_method()
        self.patch('homekeeper.common')
        self.patch('homekeeper.core')
        self.config = homekeeper.config.Config()

    def setup_symlink(self, os):
        source = self.setup_file(self.dotfiles_directory, '.vimrc')
        target = os.path.join(self.home, '.vimrc')
        return source, target

    def verify_symlink(self, os, source, target): # pylint: disable=no-self-use
        assert os.path.exists(source)
        assert os.path.exists(target)
        assert os.path.islink(target)
        assert source == os.readlink(target)
        if os.path.isfile(source):
            assert os.path.isfile(target)
        if os.path.isdir(source):
            assert os.path.isdir(target)

    def test_directory_symlinks(self, os):
        source = self.setup_directory(self.dotfiles_directory, '.vim')
        target = self.setup_directory(self.home, '.vim')
        core.symlink(self.config, source, target)
        self.verify_symlink(os, source, target)

    def test_symlink_with_no_target(self, os):
        source, target = self.setup_symlink(os)
        core.symlink(self.config, source, target)
        self.verify_symlink(os, source, target)

    def test_symlink_with_file_target(self, os):
        source, target = self.setup_symlink(os)
        self.setup_file(target)
        self.config.overwrite = True
        core.symlink(self.config, source, target)
        self.verify_symlink(os, source, target)

    def test_symlink_with_directory_target(self, os):
        source, target = self.setup_symlink(os)
        self.setup_directory(os.path.dirname(target))
        self.config.overwrite = True
        core.symlink(self.config, source, target)
        self.verify_symlink(os, source, target)

    def test_symlink_with_symlink_target(self, os):
        source, target = self.setup_symlink(os)
        original_source = self.setup_file(self.home, 'vimrc')
        os.symlink(original_source, target)
        self.config.overwrite = True
        core.symlink(self.config, source, target)
        self.verify_symlink(os, source, target)

    def test_symlink_with_no_overwrite(self, os):
        source, target = self.setup_symlink(os)
        self.setup_file(target)
        self.config.overwrite = False
        core.symlink(self.config, source, target)
        assert os.path.exists(source)
        assert os.path.exists(target)
        assert not os.path.islink(target)

    def test_symlink_with_same_source_and_target(self, os):
        source, target = self.setup_symlink(os)
        self.config.overwrite = True
        core.symlink(self.config, source, source)
        assert not os.path.exists(target)
        assert not os.path.islink(source)

    def test_restore_file_symlink(self, os):
        source, target = self.setup_symlink(os)
        core.symlink(self.config, source, target)
        core.restore(self.config, source, target)
        assert not os.path.islink(target)
        assert os.path.isfile(target)

    def test_restore_directory_symlink(self, os):
        source = os.path.join(self.dotfiles_directory, '.vim')
        target = os.path.join(self.home, '.vim')
        self.setup_file(source, '.vim', 'autoload', 'pathogen.vim')
        core.symlink(self.config, source, target)
        core.restore(self.config, source, target)
        assert not os.path.islink(target)
        assert os.path.isdir(target)
        assert os.path.isfile(os.path.join(target, '.vim', 'autoload',
                                           'pathogen.vim'))

    def test_restore_symlink_with_no_target(self, os):
        source, target = self.setup_symlink(os)
        core.symlink(self.config, source, target)
        os.unlink(target)
        core.restore(self.config, source, target)
        assert os.path.exists(source)
        assert not os.path.exists(target)

    def test_restore_symlink_with_same_source_and_target(self, os):
        source, target = self.setup_symlink(os)
        core.symlink(self.config, source, target)
        core.restore(self.config, target, target)
        assert os.path.islink(target)
        assert source == os.readlink(target)

    def test_create_symlinks(self, os):
        source, target = self.setup_symlink(os)
        source_directory = os.path.dirname(source)
        assert os.path.exists(source)
        assert not os.path.exists(target)
        self.config.overwrite = True
        core.create_symlinks(self.config, source_directory, self.home)
        self.verify_symlink(os, source, target)

    def test_create_symlinks_with_same_source_and_target(self, os):
        core.create_symlinks(self.config, self.home, self.home)
        for item in os.listdir(self.home):
            assert not os.path.islink(os.path.join(self.home, item))

    def test_create_symlinks_with_no_source_directory(self, os):
        self.setup_directory(self.home)
        source_directory = os.path.join(self.home, 'non-existant-directory')
        core.create_symlinks(self.config, source_directory, self.home)
        for item in os.listdir(self.home):
            assert not os.path.islink(item)

    def test_create_symlinks_with_no_overwrite(self, os):
        source, target = self.setup_symlink(os)
        source_directory = os.path.dirname(source)
        self.setup_file(target)
        self.config.overwrite = False
        core.create_symlinks(self.config, source_directory, self.home)
        assert os.path.exists(source)
        assert os.path.exists(target)
        assert not os.path.islink(target)

    def test_create_symlinks_with_multiple_sources(self, os):
        source_files = ['.bash_profile', '.gitignore', '.gvimrc', '.vimrc']
        source_directories = ['bin', '.git', '.vim']
        self.config.excludes = ['.git', '.gitignore']
        self.config.overwrite = True
        for item in source_files:
            self.setup_file(self.dotfiles_directory, item)
            self.setup_file(self.home, item)
        for item in source_directories:
            self.setup_directory(self.dotfiles_directory, item)
            self.setup_file(self.home, item)
        core.create_symlinks(self.config, self.dotfiles_directory, self.home)
        for item in source_files:
            if item in self.config.excludes:
                assert not os.path.islink(os.path.join(self.home, item))
            else:
                source = os.path.join(self.dotfiles_directory, item)
                target = os.path.join(self.home, item)
                self.verify_symlink(os, source, target)

    def test_create_symlinks_with_includes(self, os):
        include = self.setup_file(self.dotfiles_directory, '.foo', 'bar',
                                  'bazrc')
        target = os.path.join(self.home, '.foo', 'bar', 'bazrc')
        self.config.includes = [include]
        core.create_symlinks(self.config, self.dotfiles_directory, self.home)
        assert not os.path.islink(os.path.join(self.home, '.foo'))
        assert os.path.exists(target)
        assert os.path.islink(target)
        assert os.readlink(target) == include

    def test_create_symlinks_with_single_directory_includes(self, os):
        include = self.setup_file(self.dotfiles_directory, '.foo')
        target = os.path.join(self.home, '.foo')
        self.config.includes = [include]
        core.create_symlinks(self.config, self.dotfiles_directory, self.home)
        assert os.path.exists(target)
        assert os.path.islink(target)
        assert os.readlink(target) == include

    def test_create_with_bad_includes(self, os):
        self.config.includes = ['', 'non-existing', os.sep]
        core.create_symlinks(self.config, self.dotfiles_directory, self.home)
        assert not os.path.exists(os.path.join(self.dotfiles_directory,
                                               'non-existing'))

    def test_cleanup_symlinks(self, os):
        with cd(self.home):
            self.setup_file('existing.txt')
            os.symlink('existing.txt', 'existing-link.txt')
            os.symlink('non-existing-1.txt', 'non-existing-1-link.txt')
            os.symlink('non-existing-2.txt', 'non-existing-2-link.txt')
            core.cleanup_symlinks(self.home)
            assert os.path.exists('existing.txt')
            assert os.path.exists('existing-link.txt')
            assert not os.path.exists('non-existing-1-link.txt')
            assert not os.path.exists('non-existing-2-link.txt')

    def test_firstpart(self, os):
        assert core.firstpart(os.path.join(os.sep, 'foo', 'bar')) == os.sep
        assert core.firstpart(os.path.join('foo', 'bar')) == 'foo'
        assert core.firstpart('bar') == 'bar'

    def test_relativize(self, os):
        base_directory = os.path.join(os.sep, 'foo', 'bar', 'baz')
        assert core.relativize(os.path.join(base_directory, 'bif'),
                               base_directory) == 'bif'

    def test_is_valid_include(self, os):
        commonprefix = self.dotfiles_directory
        os.chdir(commonprefix)
        assert core.is_valid_include(os.sep, commonprefix) == False
        assert core.is_valid_include('foobar', commonprefix) == False
        assert core.is_valid_include('..', commonprefix) == False
        assert core.is_valid_include('.', commonprefix) == False
        self.setup_file(self.dotfiles_directory, 'foobar')
        assert core.is_valid_include('foobar', commonprefix) == True



import homekeeper
import setuptools

setuptools.setup(
    author='Min Huang',
    author_email='min.huang@alumni.usc.edu',
    description='Homekeeper can version your dotfiles with Git.',
    long_description=open('README.txt').read(),
    name='homekeeper',
    packages=['homekeeper'],
    install_requires=['click >= 6.7'],
    version=homekeeper.__version__,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Utilities',
    ],
    entry_points="""
        [console_scripts]
            homekeeper = homekeeper.cli:main
    """,
)
